
from Tkinter import *
import time
from threading import Thread, Timer

class SpeechRecognition(Thread):
    
    WORD_TIMEOUT_TIME = 3
    SENTENCE_TIMEOUT_TIME = 1
    WORD_END_LIST = [" ",".","?","!"]
    SENTENCE_END_LIST = [".","?","!"]
    VALID_CHARACTER_LIST = ["a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z"
                            ,"A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z"
                            ," ",".",",",":",";","`","~","?","!"
                            ,"1","2","3","4","4","5","6","7","8","9","0"
                            ]
    
    CHARACTER_ONLY_LIST = ["a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z"
                            ,"A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z"
                            ]
    
    # standAlone =  True creates own inputField
    #inputField class with set/get text method
    #            and setKeyHandler function
    def __init__(self,standAlone=True,inputField=None):
        Thread.__init__(self)
        self.standAlone = standAlone
        self.inputField = inputField
        
        #set variables
        self.sentenceEvent = []
        self.wordEvent = []
        self.characterEvent = []
        self.keyWordRecognition = False
        self.keyWordRecognitionList = []
        self.keyWordRecognitionFunction = None
        self.activeWord = Internal_Word(self.WORD_END_LIST)
        self.activeSentence = Internal_Sentence()
        
        
        if standAlone:
            #create own input field
            self.root = Tk()
            self.root.title("SpeechRecognition Input Field")
            self.inputField = SpeechRecognitionInputField(self.root)
                       
            
            self.inputField.grid(row=0,column=0)
            self.inputField.focus_force()
            
            self.root.protocol("WM_DELETE_WINDOW", self.quit)
        
        self.inputField.setKeyHandler(self.characterTyped)
        
        Thread.start(self)
    
    #Internal function
    def run(self):
        #self.root.after(100, self.show)
        self.timer = Timer(0.1, self.show)
        self.timer.start()
        if self.standAlone:
            self.root.mainloop()
        else:
            pass 
    
    #Call this function for a clean exit of the program
    def quit(self):
        
        self.timer.cancel()
        self.root.destroy()
    
    #Interal function
    #Handles the tick mechanisme
    def show(self):
        self.timer = Timer(0.1, self.show)
        self.timer.start()
        self.checkActivity()
    
    #Auto close words and sentences 
    def checkActivity(self):
        
        timeNow = time.time()
        
        if not self.activeWord.getEndTime() == None and timeNow - self.activeWord.getEndTime() >= self.WORD_TIMEOUT_TIME:
            self.addCharacterToActiveWord(Internal_Character(self.WORD_END_LIST[0],timeNow))
        
        if not self.activeSentence.getEndTime() == None and timeNow - self.activeSentence.getEndTime() >= self.SENTENCE_TIMEOUT_TIME and self.activeWord.getEndTime() == None:
            self.setNewActiveSentence()
    
    #Time in seconds 
    #Time untill a word is closed after no input        
    def setWordTimoutTime(self,time):
        self.WORD_TIMEOUT_TIME = time
    
    #Time in seconds 
    #Time untill a sentence is closed after no input
    #wait's untill last word is closed    
    def setSentenceTimeoutTime(self,time):
        self.SENTENCE_TIMEOUT_TIME = time
   
    #Set list with characters noting the end of a word
    # !!! All characters in the list should also be in the validCharacterList
    #Default list can be found at the start of this file
    def setWordEndList(self,list):
        self.WORD_END_LIST = list
        self.activeWord = Internal_Word(self.WORD_END_LIST)
    
    #Add a character to the wordEndList
    #Also add character to validCharacterList
    #Default list can be found at the start of this file
    def addWordEnd(self,char):
        self.WORD_END_LIST.append(char)
        self.activeWord = Internal_Word(self.WORD_END_LIST)
        self.VALID_CHARACTER_LIST.append(char)
        
    #Set list with characters noting the end of a Sentence
    # !!! all characters in sentenceEndList should also be a validCharacter and wordEndCharacter
    #Default list can be found at the start of this file    
    def setSentenceEndList(self,list):
        self.SENTENCE_END_LIST = list
    
    #Add a character to the sentenceEndList
    #Also add the character to the wordEndList because sentence end is also a word end
    #Also add the character to the valid character list
    #Default list can be found at the start of this file    
    def addSentenceEnd(self,char):
        self.SENTENCE_END_LIST.append(char)
        self.WORD_END_LIST.append(char)
        self.VALID_CHARACTER_LIST.append(char)
    
    #Set list as validCharacterList
    #Default list can be found at the start of this file    
    def setValidCharacterList(self,list):
        self.VALID_CHARACTER_LIST = list
    
    #Add a character to the validCharacterList
    #Default list can be found at the start of this file
    def addValidCharacter(self,char):
        self.VALID_CHARACTER_LIST.append(char)
    
    #Sets the characterOnlyList as validCharacterList
    #Default list can be found at the start of this file    
    def setCharactersOnly(self):
        self.setValidCharacterList(self.CHARACTER_ONLY_LIST)
    
    #Main input function
    #All read text should be entered character by character to this function
    #this function is auto registerd to valid inputField with setKeyHandler function
    def characterTyped(self,character):
        
        try:
            self.VALID_CHARACTER_LIST.index(character)
            charObject = Internal_Character(character,time.time())
            self.addCharacterToActiveWord(charObject)
                   
            self.inputField.set("")
            
            for function in self.characterEvent:
                function(charObject)
        except:
            
            self.inputField.set("")
            self.printMessage("invalid character: "+str(character))
    
    #Internal function
    #Adds read character to the current active word
    #closes word and activates wordEvent functions      
    def addCharacterToActiveWord(self,charObject):
        if self.activeWord.addCharacter(charObject):
            self.activeSentence.addWord(self.activeWord)
           
           
            for function in self.wordEvent:
                function(self.activeWord)
               
            self.activeWord = Internal_Word(self.WORD_END_LIST)  
            try:
                self.SENTENCE_END_LIST.index(charObject.getCharacter())
                self.setNewActiveSentence()
               
            except:
                pass
    
    #Add sentenceEvent handling function
    #function is called when a end of sentence event is found 
    #format: function(internal_sentence)
    def addSentenceEvent(self,function):
        try:
            self.sentenceEvent.index(function)
            self.printMessage("Trying to add same function twice")
        except:
            self.sentenceEvent.append(function)
    
    #Add wordEvent handling function 
    #function is called when a end of word event is found
    #format: function(Internal_Word)        
    def addWordEvent(self,function):
        try:
            self.wordEvent.index(function)
            self.printMessage("Trying to add same function twice")
        except:
            self.wordEvent.append(function)
    
    #Add characterEvent handling function
    #function is called when a character event is found    
    def addCharacterEvent(self,function):
        try:
            self.characterEvent.index(function)
            self.printMessage("Trying to add same function twice")
        except:
            self.characterEvent.append(function)      
    
    #Internal function
    #sets a new active sentence and activates all sentenceEvent handlers
    #format: function(internal_sentence)
    def setNewActiveSentence(self):
        sentence = self.activeSentence
        self.activeSentence = Internal_Sentence()
        for function in self.sentenceEvent:
            try:
                function(sentence)
            except:
                pass
        
    
    
    #Internal function                
    def printMessage(self,message):
        print message

class KeyWordDetection():
    
    def __init__(self,speechRecognizer):
        self.SR = speechRecognizer
        self.keyWordRecognitionList = []
        self.pause = False
        self.SR.addWordEvent(self.wordRecognition)
    
    #set keyWordRecognition function
    #function is called everytime a keyWord is found
    #format: function(Internal_Word,stringWord,foundStringWord,SR_Word)
    def setKeyWordRecognition(self,function,sr_wordList):
        if not function == None:        
            self.keyWordRecognitionList = []
            self.keyWordRecognitionFunction = function
            
            for word in sr_wordList:
                word.setValidCharacters(self.SR.VALID_CHARACTER_LIST)
                self.keyWordRecognitionList.append(word)
        else:
            self.keyWordRecognitionFunction = None
            self.keyWordRecognitionList = []
            
    #KeyWordRecognition needs to be set with setKeyWordRecognition
    def addKeyWordRecognitionWord(self,keyWord):
        if not self.keyWordRecognitionFunction == None:
            keyWord.setValidCharacters(self.SR.VALID_CHARACTER_LIST)
            self.keyWordRecognitionList.append(keyWord)
    
    #Internal function
    #word event handler used by the keyWordRecognition        
    def wordRecognition(self,wordObject):
        if not self.pause:            
            for sr_word in self.keyWordRecognitionList:
                
                index  = sr_word.indexOf(wordObject.getWord())
                if index >= 0:    
                    self.keyWordRecognitionFunction(wordObject,sr_word.getWord(),sr_word.get(index),sr_word)
        
    def pauseRun(self):
        self.pause = True
        
    def continueRun(self):
        self.pause = False
    
    def clear(self):
        self.keyWordRecognitionList = []
            
    def quit(self):
        self.SR.quit()    
    
class KeySentenceDetection():
    
    def __init__(self,speechRecognition):
        self.SR = speechRecognition
        self.KR = KeyWordRecognition(self.SR)
        self.sentenceRecognitionFunction = None
        self.sentenceRecognitionThreshold = 0
        self.sentenceTuples = []
    
    # set sentenceRecognition function
    # function is called everytime a sentence end event is found
    # function get's None if no sentence has been recognized above threshold certainty
    # sentences are compared character based so long words recognized have greater impact
    # then smaller words
    # format: function(sr_sentence)    
    def setSentenceRecognition(self,function,sr_sentenceList,threshold):
        if not function == None:
            sr_wordList = []
            self.sentenceList = dict()
            self.sentenceRecognitionFunction = function
            self.sentenceRecognitionThreshold = threshold
            self.sentenceTuples = []
            for sr_sentence in sr_sentenceList:
                   
                tuple = (sr_sentence,[])
                for sr_word in sr_sentence.getWords():
                    sr_wordList.append(sr_word)
                    try:
                        list = self.sentenceList[sr_word.getWord().upper()]
                        list.append(tuple)                       
                        self.sentenceList[sr_word.getWord().upper()] = list
                        
                    except:
                        self.sentenceList[sr_word.getWord().upper()] = [tuple]            
                
                self.sentenceTuples.append(tuple)
                
            self.KR.setKeyWordRecognition(self.sentenceRecognition, sr_wordList)
            
        else:
            self.clear()            
            self.sentenceRecognitionFunction = function
            self.sentenceRecognitionThreshold = 0
                        
    
    #Internal function
    #used for the sentenceRecognition code  
    #word end event handler      
    def sentenceRecognition(self,wordObject,word,detectedWord,sr_word):
        
        if len(self.sentenceList) > 0:
        
            tuples = self.sentenceList[sr_word.getWord().upper()]
            for tuple in tuples:
                try:
                    tuple[1].index(sr_word.getWord().upper())
                except:
                    
                    tuple[1].append(sr_word.getWord().upper()) 
    
    #Internal function
    #used for the sentenceRecognition code
    #sentence end event handler                
    def sentenceRecognitionFinish(self,sentenceObject):
        #sentence that got over threshold
        #print "sentence regocnition finish "+str(sentenceObject)
        recognized = []
        
        for tuple in self.sentenceTuples:
            #print "step1 "+str(tuple)
            sr_sentence = tuple[0]
            total = 0.0
            for sr_word in sr_sentence.getWords():
                total += len(sr_word.getWord())
                
            found =  tuple[1]
            counter = 0.0
            #print "step2"
            for word in found:
                counter += len(word)
                
            
            if counter / total >= self.sentenceRecognitionThreshold:
                recognized.append((sr_sentence,counter/total))
            #print "step3"
        if len(recognized) > 0:
            #print "step4"
            top = recognized[0]
            for contender in recognized:
                if len(top[0].getWords()) <= 2:
                    top = contender
                elif top[1] < contender[1]:
                    top = contender
            #print "step5"
            self.sentenceRecognitionFunction(top[0])
                    
        else:
            self.sentenceRecognitionFunction(None)
            
    def pause(self):
        self.KR.pause()
    
    def continueRun(self):
        self.KR.continueRun()
        
    def clear(self):
        self.KR.clear()
        self.sentenceList = Dict()
        self.sentenceTuples = []    

#Example of correct inputField class            
class SpeechRecognitionInputField(Entry):
    
    #Tkinter type Entry field for easy SpeechRecognition input
    def __init__(self,parent):
        
        Entry.__init__(self,parent)
        self.input = StringVar()
        self.config(textvariable=self.input)
        self.bind("<Key>",self.handleKey )
    
    #Get the text from input field    
    def get(self):
        return self.input.get()
    
    #Set the text in the input field
    def set(self,string):
        return self.input.set(string)
    
    #Set keyHandler function 
    #used to integrade SpeechRecognition code with the field
    def setKeyHandler(self,function):
        self.characterTyped = function
    
    #auto call the keyHandler function of the SpeechRecognition   
    def handleKey(self,event):
        self.characterTyped(event.char)
        
class Internal_Character:
        
    def __init__(self,char,time):
        self.character = char
        self.time = time
        
    def getCharacter(self):
        return self.character
    
    def getTime(self):
        return self.time
    
    def __str__(self):
        return self.getCharacter()
    
class Internal_Word:
    
    def __init__(self,wordEnds):
        self.wordList = []
        self.startTime = None
        self.endTime = None
        self.word = ""
        self.wordEnds = wordEnds
        
    def addCharacter(self,charObject):
        if self.startTime == None:
            try:
                self.wordEnds.index(charObject.getCharacter())
            except:
                self.startTime = charObject.getTime()
        
        self.wordList.append(charObject)
        
        self.endTime = charObject.getTime()
        
        try:
            self.wordEnds.index(charObject.getCharacter())
            return True
        except:        
            self.word += charObject.getCharacter()
            return False
        
    def getStartTime(self):
        return self.startTime
    
    def getEndTime(self):
        return self.endTime
    
    def getWord(self,validCharacters=[]):
        return self.word
    
    def getWordList(self):
        return self.wordList
    
    def getWordDuration(self):
        return self.endTime - self.startTime
    
    def __str__(self):
        return self.getWord()

class Internal_Sentence:
    
    def __init__(self):
        self.sentenceList = []
        self.startTime = None
        self.endTime = None
        self.sentence = ""
        
    def addWord(self,wordObject):
        if self.startTime == None:
            self.startTime = wordObject.getStartTime()
        
        self.sentenceList.append(wordObject)
        
        self.endTime = wordObject.getEndTime()
        
        self.sentence += wordObject.getWord()+" "
            
    def getSentence(self):
        return self.sentence
    
    def getStartTime(self):
        return self.startTime
    
    def getEndTime(self):
        return self.endTime
    
    def getSentenceList(self):
        return self.sentenceList
    
    def getSentenceDuration(self):
        return self.endTime - self.startTime
    
    def __str__(self):
        return self.getSentence()

class SR_Word():

    def __init__(self,word,alternatives = [],key=None):
        self.word = word
        self.alternatives = alternatives
        self.key = key
        self.validCharacters = []
    
    def setAlternatives(self,alternatives):
        self.alternatives = alternatives
        
    def addAlternatives(self,alternatives):
        self.alternatives.extend(alternatives)
        
    def addAlternative(self,alternative):
        self.alternatives.append(alternative)
    
    def getWord(self):
        return self.word
    
    def getAlternatives(self):
        return self.alternatives
    
    def getKey(self):
        return self.key
    
    def setValidCharacters(self,validCharacters):
        self.validCharacters = validCharacters
        
        tempWord = ""
        
        for letter in self.word:
            try:
                self.validCharacters.index(letter)
                tempWord += letter
            except:
                pass
        
        self.word = tempWord
        
        
        tempAlternatives = []
        for alternative in self.alternatives:
            tempAlternative = ""
            for letter in alternative:
                try:
                    self.validCharacters.index(letter)
                    tempAlternative += letter
                except:
                    pass
            
            tempAlternatives.append(tempAlternative)
        self.alternatives = tempAlternatives
        
    def indexOf(self,wordString):
        upWord = self.word.upper()
        upWordString = wordString.upper()
        
        if upWord == upWordString:
            return 0
        
        else:
            counter = 0
            while counter < len(self.alternatives):
                upAlternative = self.alternatives[counter].upper()
                if upAlternative == upWordString:
                    return counter + 1
                
                counter += 1
                
        return -1
    
    def get(self,index):
        if index == 0:
            return self.word
        
        else:
            if len(self.alternatives) > index-1:
                return self.alternatives[index-1]
        
    def __str__(self):
        alter = ""
        for alternative in self.alternatives:
            alter += " "+str(alternative)
    
        return str(self.word)+":"+alter
    
    
    
class SR_Sentence():
    
    def __init__(self,sentence="",key=None):
        self.wordList = []
        self.key = key
        self.sentence = sentence
        
        if len(sentence) > 0:
            sentenceList = sentence.split()
            for word in sentenceList:
                self.wordList.append(SR_Word(word))
        
    def addWord(self,sr_word):
        self.wordList.append(sr_word)
    
    def addWordList(self,sr_wordList):    
        self.wordList.extend(sr_wordList)
        
    def setSentence(self,sr_wordList):
        self.wordList = sr_wordList
    
    def getSentence(self):
        if self.sentence == None:
            return str(self)
        else:
            return self.sentence
    
    def getWords(self):
        return self.wordList
    
    def getKey(self):
        return self.key
    
    def __str__(self):
        sentence = ""
        
        for sr_word in self.wordList:
            sentence += sr_word.getWord()+" "
            
        return sentence




    
if __name__ == '__main__':
    SR = SpeechRecognition()